package it.inaf.oats.ia2.tapschemamanager.datalayer;

import java.io.Serializable;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;
import javax.persistence.EntityManager;
import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.CriteriaQuery;

class SchemaInfo implements Serializable {
    
    private static final long serialVersionUID = 2835388936102349727L;

    private final List<String> allSchemaTables;
    private final List<KeyEntity> schemaKeys;
    private final Map<String, String> tablesType;

    SchemaInfo(Connection connection, String schemaName) throws SQLException {
        allSchemaTables = DataProvider.getAllTablesNames(connection, schemaName);
        schemaKeys = DataProvider.getSchemaKeys(connection, schemaName);
        tablesType = DataProvider.getTablesTypes(connection, schemaName);
    }

    public List<String> getAllSchemaTables() {
        return allSchemaTables;
    }

    public List<KeyEntity> getSchemaKeys() {
        return schemaKeys;
    }

    public Map<String, String> getTablesType() {
        return tablesType;
    }
}

class ColumnsInfo implements Serializable {
    
    private static final long serialVersionUID = -3407052226160044192L;

    private final Map<String, Map<String, List<ColumnInfo>>> columnsInfo;

    ColumnsInfo() {
        this.columnsInfo = new HashMap<String, Map<String, List<ColumnInfo>>>();
    }

    void addColumnsInfo(Connection connection, TableEntity tableEntity) throws SQLException {
        String schemaName = tableEntity.getSchemaName();
        Map<String, List<ColumnInfo>> map = columnsInfo.get(schemaName);
        if (map == null) {
            map = new HashMap<String, List<ColumnInfo>>();
            columnsInfo.put(schemaName, map);
        }
        map.put(tableEntity.getShortTableName(), DataProvider.getAllColumns(connection, tableEntity));
    }

    List<ColumnInfo> getColumnsInfo(String schemaName, String shortTableName) {
        return columnsInfo.get(schemaName).get(shortTableName);
    }
}

/**
 *
 * @author Sonia Zorba <zorba at oats.inaf.it>
 */
public class TapSchemaHandler implements Serializable {
    
    private static final long serialVersionUID = 1193743698810532785L;

    private final Credentials credentials;
    private final String name;
    private boolean exists;

    private final List<String> allSchemas;
    private final List<SchemaEntity> schemas;
    private final List<String> persisedSchemas;
    private final List<SchemaEntity> toRemoveSchemas;

    private final Map<String, SchemaInfo> allSchemaInfos;

    private final ColumnsInfo columnsInfo;

    private transient EntityManager entityManager;

    private transient Connection connection;

    public TapSchemaHandler(Credentials credentials, String tapSchemaName, boolean exists) throws SQLException {
        this.credentials = credentials;
        this.name = tapSchemaName;
        this.exists = exists;

        Connection conn = getConnection();

        try {
            this.allSchemas = DataProvider.getAllSchemasNames(conn);

            this.schemas = new ArrayList<SchemaEntity>();
            this.persisedSchemas = new ArrayList<String>();
            this.toRemoveSchemas = new ArrayList<SchemaEntity>();

            this.allSchemaInfos = new HashMap<String, SchemaInfo>();

            this.columnsInfo = new ColumnsInfo();

            if (exists) {
                entityManager = DataProvider.getEntityManager(credentials, tapSchemaName);
                CriteriaBuilder cb = entityManager.getCriteriaBuilder();
                CriteriaQuery<SchemaEntity> cq = cb.createQuery(SchemaEntity.class);
                for (SchemaEntity schemaEntity : entityManager.createQuery(cq).getResultList()) {
                    String schemaName = schemaEntity.getName();
                    addSchema(schemaEntity);
                    allSchemaInfos.put(schemaName, new SchemaInfo(conn, schemaName));
                    persisedSchemas.add(schemaName);
                    for (TableEntity tableEntity : schemaEntity.getTables()) {
                        columnsInfo.addColumnsInfo(conn, tableEntity);
                    }
                }
            }

        } catch (SQLException e) {
            throw e;
        } finally {
            closeConnection();
        }
    }

    public SchemaEntity getNewSchema(String schemaName) throws SQLException {
        SchemaEntity schemaEntity = new SchemaEntity(schemaName);

        Connection conn = getConnection();
        try {
            allSchemaInfos.put(schemaName, new SchemaInfo(conn, schemaName));
        } catch (SQLException e) {
            throw e;
        } finally {
            closeConnection();
        }

        return schemaEntity;
    }

    public final void addSchema(SchemaEntity schemaEntity) {
        DLUtil.insertEntityInOrder(schemas, schemaEntity);
    }

    public final void removeSchema(String schemaName) {
        if (persisedSchemas.contains(schemaName)) {
            toRemoveSchemas.add(getSchema(schemaName));
            persisedSchemas.remove(schemaName);
        }
        DLUtil.removeEntity(schemas, schemaName);
    }

    public List<SchemaEntity> getSchemas() {
        return schemas;
    }

    public List<String> getAllSchemas() {
        return allSchemas;
    }

    public List<String> getAllTables(String schemaName) {
        return allSchemaInfos.get(schemaName).getAllSchemaTables();
    }

    public List<ColumnInfo> getColumnInfo(String schemaName, String shortTableName) {
        return columnsInfo.getColumnsInfo(schemaName, shortTableName);
    }

    public int getMaxKey() {
        int maxKey = 0;
        for (SchemaEntity schemaEntity : schemas) {
            for (TableEntity tableEntity : schemaEntity.getTables()) {
                for (KeyEntity keyEntity : tableEntity.getFromKeys()) {
                    int key = Integer.parseInt(keyEntity.getKeyId());
                    if (key > maxKey) {
                        maxKey = key;
                    }
                }
                for (KeyEntity keyEntity : tableEntity.getTargetKeys()) {
                    int key = Integer.parseInt(keyEntity.getKeyId());
                    if (key > maxKey) {
                        maxKey = key;
                    }
                }
            }
        }
        return maxKey;
    }

    public TableEntity getNewTable(String schemaName, String shortTableName) throws SQLException {
        TableEntity table = new TableEntity(getSchema(schemaName), shortTableName);

        table.setTableType(allSchemaInfos.get(schemaName).getTablesType().get(shortTableName));
        
        Connection conn = getConnection();
        try {
            // Add columns info
            // For performance reasons all possible column entities are loaded together
            columnsInfo.addColumnsInfo(conn, table);
        } catch (SQLException e) {
            throw e;
        } finally {
            closeConnection();
        }

        return table;
    }

    public void addTable(String schemaName, TableEntity tableEntity) {
        SchemaEntity schema = getSchema(schemaName);
        schema.addTable(tableEntity);

        // Add table keys
        int currentKey = getMaxKey() + 1;
        for (KeyEntity keyEntity : getAllSchemaKeys(schemaName)) {
            String fromTableName = keyEntity.getFromTableFullName();
            String targetTableName = keyEntity.getTargetTableFullName();
            TableEntity fromTable = null, targetTable = null;
            if (tableEntity.getName().equals(fromTableName)) {
                fromTable = tableEntity;
                targetTable = schema.getTable(targetTableName);
            } else if (tableEntity.getName().equals(targetTableName)) {
                targetTable = tableEntity;
                fromTable = schema.getTable(fromTableName);
            }
            if (fromTable != null && targetTable != null
                    && !fromTable.getFromKeys().contains(keyEntity)
                    && !targetTable.getTargetKeys().contains(keyEntity)) {
                String key = currentKey + "";
                keyEntity.setKeyId(key);
                keyEntity.setFromTable(fromTable);
                keyEntity.setTargetTable(targetTable);
                for (KeyColumnEntity keyColumn : keyEntity.getKeyColumns()) {
                    keyColumn.setKeyId(key);
                }

                fromTable.addFromKey(keyEntity);
                targetTable.addTargetKey(keyEntity);

                currentKey++;
            }
        }
    }

    public void removeTable(SchemaEntity schemaEntity, TableEntity tableEntity) {
        for (KeyEntity keyEntity : tableEntity.getFromKeys()) {
            keyEntity.getTargetTable().removeTargetKey(keyEntity);
        }
        for (KeyEntity keyEntity : tableEntity.getTargetKeys()) {
            keyEntity.getFromTable().removeFromKey(keyEntity);
        }

        schemaEntity.removeTable(tableEntity.getName());
    }

    public void addColumn(TableEntity tableEntity, ColumnEntity columnEntity) {
        tableEntity.addColumn(columnEntity);
    }

    public void removeColumn(TableEntity tableEntity, String columnName) {
        tableEntity.removeColumn(columnName);
    }

    public String getForeignKeyReference(String schemaName, String fullTableName, String columnName) {
        for (KeyEntity key : getAllSchemaKeys(schemaName)) {
            if (key.getFromTableFullName().equals(fullTableName)) {
                for (KeyColumnEntity keyColumn : key.getKeyColumns()) {
                    if (keyColumn.getFromColumn().equals(columnName)) {
                        return key.getTargetTableFullName().split(Pattern.quote("."))[1] + "." + keyColumn.getTargetColumn();
                    }
                }
            }
        }

        return null;
    }

    public SchemaEntity getSchema(String schemaName) {
        return DLUtil.getEntityByName(schemas, schemaName);
    }

    private Connection getConnection() throws SQLException {
        if (connection == null) {
            connection = credentials.getConnection();
        }
        return connection;
    }

    private void closeConnection() throws SQLException {
        if (connection != null) {
            try {
                connection.close();
            } catch (SQLException e) {
                throw e;
            } finally {
                connection = null;
            }
        }
    }

    public List<KeyEntity> getAllSchemaKeys(String schemaName) {
        return allSchemaInfos.get(schemaName).getSchemaKeys();
    }

    public void save() throws SQLException {
        if (exists) {
            entityManager.getTransaction().begin();
            for (SchemaEntity toRemoveSchema : toRemoveSchemas) {
                entityManager.remove(toRemoveSchema);
            }
            for (SchemaEntity schemaEntity : schemas) {
                entityManager.merge(schemaEntity);
            }
            entityManager.getTransaction().commit();
        } else {
            try {
                DataProvider.createDatabase(getConnection(), name);
            } catch (SQLException e) {
                throw e;
            } finally {
                closeConnection();
            }
            entityManager = DataProvider.getEntityManager(credentials, name);

            entityManager.getTransaction().begin();
            for (SchemaEntity schemaEntity : schemas) {
                entityManager.persist(schemaEntity);
            }
            entityManager.getTransaction().commit();

            // Adding TAP_SCHEMA into TAP_SCHEMA
            SchemaEntity tapSchemaSchema = getNewSchema(name);
            addSchema(tapSchemaSchema);

            for (String shortTableName : allSchemaInfos.get(name).getAllSchemaTables()) {
                TableEntity tableEntity = getNewTable(name, shortTableName);
                addTable(name, tableEntity);
                for (ColumnInfo ci : columnsInfo.getColumnsInfo(name, shortTableName)) {
                    addColumn(tableEntity, ci.getColumnEntity());
                }
            }

            DLUtil.putInfoIntoTapSchemaSchema(tapSchemaSchema);

            entityManager.getTransaction().begin();
            entityManager.merge(tapSchemaSchema);
            entityManager.getTransaction().commit();

            exists = true;
        }
    }
}
